"""Interface with the Replit Database."""
from .database import AsyncDatabase, Database, DBJSONEncoder, dumps, to_primitive
from .default_db import db, db_url
from .server import make_database_proxy_blueprint, start_database_proxy

__all__ = [
    "AsyncDatabase",
    "Database",
    "db",
    "DBJSONEncoder",
    "db_url",
    "dumps",
    "make_database_proxy_blueprint",
    "start_database_proxy",
    "to_primitive",
]



# flake8: noqa
"""Make apps quickly using Python."""

import os

import flask
from werkzeug.local import LocalProxy

from .app import debug, ReplitAuthContext, run
from .user import User, UserStore
from .utils import *
from ..database import AsyncDatabase, Database, db

auth = LocalProxy(lambda: ReplitAuthContext.from_headers(flask.request.headers))



"""Utilities for working with user mappings."""
from collections.abc import Mapping, MutableMapping
from typing import Any, Iterator, Optional

import flask

from .app import ReplitAuthContext
from ..database import Database, db as real_db

db: Database = real_db  # type: ignore


class User(MutableMapping):
    """A user in the database, usually initialized by UserStore."""

    __slots__ = ("username", "prefix", "_value")

    def __init__(self, username: str, prefix: str = "") -> None:
        self.username = username
        self.prefix = prefix

    def db_key(self) -> str:
        """Returns the database key for this user."""
        return self.prefix + self.username

    def set_value(self, value: str) -> None:
        """Sets the raw value of this user in the database.

        See set() and get() for a simpler dict based API.
        Will set the key `prefix + username`.

        Args:
            value (str): The value to set in the database
        """
        db[self.db_key()] = value

    def _ensure_value(self) -> Any:
        try:
            return db[self.db_key()]
        except KeyError:
            db[self.db_key()] = {}
            return db[self.db_key()]

    def set(self, key: str, val: Any) -> None:
        """Sets a key to a value for this user's entry in the database.

        Args:
            key (str): The key to set
            val (Any): The value to set it to
        """
        self[key] = val

    def get(self, key: str, default: Any = None) -> Any:
        """Gets a value from the database if it exists, otherwise returns default.

        Args:
            key (str): The key to retrieve
            default (Any): The value to return if the key does not exist

        Returns:
            Any: The value of the key or default
        """
        # We can't let MutableMapping implement this method because if it does it will
        #  return default directly instead of wrapping it in a special database method
        return self._ensure_value().get(key, default)

    def __getitem__(self, k: str) -> Any:
        return self._ensure_value()[k]

    def __setitem__(self, k: str, v: Any) -> None:
        self._ensure_value()[k] = v

    def __delitem__(self, k: str) -> None:
        del self._ensure_value()[k]

    def __iter__(self) -> Iterator[Any]:
        return iter(self._ensure_value())

    def __len__(self) -> int:
        return sum(1 for _ in self)


class UserStore(Mapping):
    """A mapping of username to keys in the replit database."""

    __slots__ = ("prefix",)

    def __init__(self, prefix: str = "") -> None:
        self.prefix = prefix

    @property
    def current(self) -> Optional[User]:
        """The user currently logged in with repl auth, None if not logged in."""
        auth = ReplitAuthContext.from_headers(flask.request.headers)
        if auth.is_authed:
            return self[auth.name]
        return None

    def _strip_prefix(self, k: str) -> str:
        return k[len(self.prefix) :]

    def __getitem__(self, name: str) -> User:
        return User(username=name, prefix=self.prefix)

    def __iter__(self) -> Iterator[str]:
        for k in db.keys():
            if k.startswith(self.prefix):
                yield self._strip_prefix(k)

    def __len__(self) -> int:
        return sum(1 for _ in self)



# flake8: noqa

"""The Replit Python module."""

from . import web
from .audio import Audio
from .database import (
    db,
    Database,
    AsyncDatabase,
    make_database_proxy_blueprint,
    start_database_proxy,
)
from .info import ReplInfo

info = ReplInfo()

# Backwards compatibility.
def clear() -> None:
    """Clear the terminal."""
    print("\033[H\033[2J", end="", flush=True)


audio = Audio()



# flake8: noqa
from dataclasses import dataclass
from pathlib import Path
from typing import Any, List
import flask

from ..database.database import ObservedDict, ObservedList, DBJSONEncoder


@dataclass
class ReplitAuthContext:
    """A dataclass defining a Replit Auth state."""

    user_id: int
    name: str
    roles: str

    @classmethod
    def from_headers(cls, headers: dict) -> Any:
        """Initialize an instance using the Replit identification headers.

        Args:
            headers (dict): A dictionary of headers received

        Returns:
            Any: An initialized class instance
        """
        return cls(
            user_id=headers.get("X-Replit-User-Id"),
            name=headers.get("X-Replit-User-Name"),
            roles=headers.get("X-Replit-User-Roles"),
        )

    @property
    def is_authenticated(self) -> bool:
        """Check whether the user is authenticated in with Replit Auth.

        Returns:
            bool: whether or not the authentication is activated.
        """
        return bool(self.name)

    @property
    def is_authed(self) -> bool:
        """Check whether the user is authenticated in with Replit Auth.

        Returns:
            bool: whether or not the authentication is activated.
        """
        return bool(self.name)


def run(
    app: flask.Flask,
    host: str = "0.0.0.0",
    port: int = 8080,
    change_encoder: bool = True,
    **kwargs
) -> None:
    """A simple wrapper around app.run() with replit compatible defaults."""
    # don't clobber user
    if change_encoder and flask.json_encoder is flask.json.JSONEncoder:
        app.json_encoder = DBJSONEncoder
    app.run(host=host, port=port, **kwargs)


def debug(
    app: flask.Flask,
    watch_dirs: List[str] = None,
    watch_files: List[str] = None,
    **kwargs: Any
) -> None:
    """Run the app in debug mode.
    Args:
        watch_dirs (List[str]): Directories whose files will be added to
            watch_files. Defaults to [].
        watch_files (List[str]): Files to watch, and if changes are detected
            the server will be restarted. Defaults to [].
        **kwargs (Any): Extra keyword arguments to be passed to the flask app's run
            method.
    """
    watch_files = list(watch_files or [])

    for directory in watch_dirs or []:
        if not isinstance(directory, Path):
            directory = Path(directory)
        watch_files += [str(f) for f in directory.iterdir() if f.is_file()]

    run(
        app,
        debug=True,
        extra_files=watch_files,
        **kwargs,
    )



"""CLI for interacting with your Repl's DB. Written as top-level script."""
import json

import click
from replit import db as database


reset = "\u001b[0m"


def info(value: str) -> str:
    """Wrap given string in a blue color for info contexts."""
    # Bright blue foreground
    return "\u001b[34;1m" + value + reset


def success(value: str) -> str:
    """Wrap given string in a green color for success contexts."""
    # Bright green foreground
    return "\u001b[32;1m" + value + reset


def failure(value: str) -> str:
    """Wrap given string in a red color for failure/warning contexts."""
    # Bright red foreground
    return "\u001b[31;1m" + value + reset


@click.group()
@click.version_option("0.0.1")
def cli() -> None:
    """CLI for interacting with your Repl's DB."""


@cli.command(name="match")
@click.argument("prefix")
def find_matches(prefix: str) -> None:
    """List all keys that match the given prefix."""
    matches = list(database.prefix(prefix))

    if matches:
        click.echo(success(f"Matches found for '{prefix}':\n"))
        click.echo("\n".join(matches))
    else:
        click.echo(failure(f"No matches found for '{prefix}'"))


@cli.command(name="set")
@click.argument("key")
@click.argument("val")
def set_value(key: str, val: str) -> None:
    """Add a given key-value pair to the DB."""
    database[key] = val
    click.echo(success(f"DB[{key}] was successfully set to '{val}'"))


@cli.command(name="del")
@click.argument("key")
def del_value(key: str) -> None:
    """Delete the key-value pair located at the given key."""
    try:
        del database[key]
    except KeyError:
        click.echo(failure(f"The key '{key}' was not found in the DB."))
    else:
        del database[key]
        click.echo(success(f"db['{key}'] was successfully deleted."))


@cli.command(name="nuke")
@click.option("--i-am-sure", is_flag=True)
def nuke_db(i_am_sure: bool) -> None:
    """Wipe ALL key-value pairs in the DB."""
    if i_am_sure:
        click.echo(info("Beginning Nuke operation...\n"))
        keys = list(database.keys())

        for k in keys:
            del database[k]

        click.echo(success("Nuke operation successful."))
    else:
        click.echo(
            failure(
                "If you REALLY want to delete everything in your database, "
                "run again with the --i-am-sure flag."
            )
        )


@cli.command(name="dump")
@click.argument("file_path")
def list_all(file_path: str) -> None:
    """Write all keys and values in the DB to a JSON file."""
    with open(file_path, "w") as f:
        keys = list(database.keys())
        binds = dict([(k, json.loads(database.get_raw(k))) for k in keys])

        json.dump(binds, f)

        click.echo(success(f"Output successfully dumped to '{file_path}'"))


if __name__ == "__main__":
    cli(prog_name="repldb")



"""A module containing a database proxy implementation."""
from typing import Any
from urllib.parse import quote

from flask import Blueprint, Flask, request

from .default_db import db


def make_database_proxy_blueprint(view_only: bool, prefix: str = "") -> Blueprint:
    """Generates a blueprint for a database proxy.

    Args:
        view_only: If False, database writing and deletion is enabled.
        prefix: A prefix that all keys interacted with using this proxy will use.

    Returns:
        Blueprint: A flask blueprint with the proxy logic.
    """
    app = Blueprint("database_proxy" + ("_view_only" if view_only else ""), __name__)

    def list_keys() -> Any:
        user_prefix = request.args.get("prefix")
        encode = "encode" in request.args
        keys = db.prefix(prefix=prefix + user_prefix)
        keys = [k[len(prefix) :] for k in keys]

        if encode:
            return "\n".join(quote(k) for k in keys)
        else:
            return "\n".join(keys)

    def set_key() -> Any:
        if view_only:
            return "Database is view only", 401
        for k, v in request.form.items():
            db[prefix + k] = v
        return ""

    @app.route("/", methods=["GET", "POST"])
    def index() -> Any:
        if request.method == "GET":
            return list_keys()
        return set_key()

    def get_key(key: str) -> Any:
        try:
            return db[prefix + key]
        except KeyError:
            return "", 404

    def delete_key(key: str) -> Any:
        if view_only:
            return "Database is view only", 401
        try:
            del db[prefix + key]
        except KeyError:
            return "", 404
        return ""

    @app.route("/<key>", methods=["GET", "DELETE"])
    def manage_key(key: str) -> Any:
        if request.method == "GET":
            return get_key(key)
        return delete_key(key)

    return app


def start_database_proxy(
    view_only: bool,
    prefix: str = "",
    host: str = "0.0.0.0",  # noqa: S104
    port: int = 8080,
) -> None:
    """Stars the database proxy."""
    app = Flask(__name__)

    app.register_blueprint(make_database_proxy_blueprint(view_only, prefix=prefix))
    app.run(host=host, port=port)



"""A module containing the default database."""
from os import environ
from typing import Optional

from .database import Database

db: Optional[Database]
db_url = environ.get("REPLIT_DB_URL")
if db_url:
    db = Database(db_url)
else:
    # The user will see errors if they try to use the database.
    db = None



"""Information about your repl."""
import os
from typing import Optional


class ReplInfo:
    """Represents info about the current repl."""

    @property
    def id(self) -> Optional[str]:
        """The id of the repl (REPL_ID environment variable)."""
        return os.getenv("REPL_ID")

    @property
    def slug(self) -> Optional[str]:
        """The slug of the repl (REPL_SLUG environement variable).

        The slug is the url-safe version of the repl's name.

        Returns:
            Optional[str]: The repl slug.
        """
        return os.getenv("REPL_SLUG")

    @property
    def owner(self) -> Optional[str]:
        """The owner of the repl (REPL_OWNER environment variable)."""
        return os.getenv("REPL_OWNER")

    @property
    def language(self) -> Optional[str]:
        """The language of the repl (REPL_LANGUAGE environment variable)."""
        return os.getenv("REPL_LANGUAGE")

    @property
    def id_co_url(self) -> Optional[str]:
        """The hosted URL of the repl in the form https://<id>.id.repl.co.

        Less readable than the vanity URL but guaranteed to work (the vanity URL might
        be too long for a certificate to be issued for it, causing it to break).

        Returns:
            Optional[str]: The id URL or None if there is no ID.
        """
        repl_id = self.id
        if repl_id is None:
            return None
        return f"https://{repl_id}.id.repl.co"

    @property
    def co_url(self) -> Optional[str]:
        """The readable, hosted repl.co URL for this repl.

        See id_url for the difference between the hosted URL types.

        Returns:
            Optional[str]: The vanity hosted URL or None if slug or owner is None.
        """
        slug = self.slug
        owner = self.owner
        if slug is None or owner is None:
            return None
        return f"https://{slug.lower()}.{owner.lower()}.repl.co"

    @property
    def replit_url(self) -> Optional[str]:
        """The URL of this repl on replit.com."""
        slug = self.slug
        owner = self.owner
        if slug is None or owner is None:
            return None
        return f"https://replit.com/@{owner}/{slug}"

    @property
    def replit_id_url(self) -> Optional[str]:
        """The URL of this repl on replit.com, based on the repl's ID."""
        repl_id = self.id
        if repl_id is None:
            return None
        return f"https://replit.com/replid/{repl_id}"



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['replit', 'replit.audio', 'replit.database', 'replit.web']

package_data = \
{'': ['*']}

install_requires = \
['Flask>=2.0.0,<3.0.0',
 'Werkzeug>=2.0.0,<3.0.0',
 'aiohttp>=3.6.2,<4.0.0',
 'requests>=2.25.1,<3.0.0',
 'typing_extensions>=3.7.4,<4.0.0']

entry_points = \
{'console_scripts': ['replit = replit.__main__:cli']}

setup_kwargs = {
    'name': 'replit',
    'version': '3.2.3',
    'description': 'A library for interacting with features of repl.it',
    'long_description': '### `>>> import replit`\n\n![compute](https://github.com/kennethreitz42/replit-py/blob/kr-cleanup/ext/readme.gif?raw=true)\n\nThis repository is the home for the `replit` Python package, which provides:\n\n- A fully-featured database client for [Replit DB](https://docs.repl.it/misc/database).\n- A Flask–based application framework for accelerating development on the platform.\n- Replit user profile metadata retrieval (more coming here!).\n- A simple audio library that can play tones and audio files!\n\n### Open Source License\n\nThis library is licensed under the [ISC License](https://en.wikipedia.org/wiki/ISC_license) and is free for you to use, change, or even profit from!\n',
    'author': 'Repl.it',
    'author_email': 'contact@repl.it',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/replit/replit-py',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Async and dict-like interfaces for interacting with Repl.it Database."""

from collections import abc
import json
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
import urllib

import aiohttp
import requests


def to_primitive(o: Any) -> Any:
    """If object is an observed object, converts to primitve, otherwise returns it.

    Args:
        o (Any): Any object.

    Returns:
        Any: The primitive equivalent if o is an ObservedList or ObservedDict,
            otherwise o.
    """
    if isinstance(o, ObservedList) or isinstance(o, ObservedDict):
        return o.value
    return o


class DBJSONEncoder(json.JSONEncoder):
    """A JSON encoder that uses to_primitive on passed objects."""

    def default(self, o: Any) -> Any:
        """Runs to_primitive on the passed object."""
        return to_primitive(o)


def dumps(val: Any) -> str:
    """JSON encode a value in the smallest way possible.

    Also handles ObservedList and ObservedDict by using a custom encoder.

    Args:
        val (Any): The value to be encoded.

    Returns:
        str: The JSON string.
    """
    return json.dumps(val, separators=(",", ":"), cls=DBJSONEncoder)


_dumps = dumps


class AsyncDatabase:
    """Async interface for Repl.it Database."""

    __slots__ = ("db_url", "sess")

    def __init__(self, db_url: str) -> None:
        """Initialize database. You shouldn't have to do this manually.

        Args:
            db_url (str): Database url to use.
        """
        self.db_url = db_url
        self.sess = aiohttp.ClientSession()

    async def __aenter__(self) -> "AsyncDatabase":
        return self

    async def __aexit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        self.sess.close()

    async def get(self, key: str) -> str:
        """Return the value for key if key is in the database.

        This method will JSON decode the value. To disable this behavior, use the
        `get_raw` method instead.

        Args:
            key (str): The key to retreive

        Returns:
            str: The the value for key if key is in the database.
        """
        return json.loads(await self.get_raw(key))

    async def get_raw(self, key: str) -> str:
        """Get the value of an item from the database.

        Args:
            key (str): The key to retreive

        Raises:
            KeyError: Key is not set

        Returns:
            str: The value of the key
        """
        async with self.sess.get(
            self.db_url + "/" + urllib.parse.quote(key)
        ) as response:
            if response.status == 404:
                raise KeyError(key)
            response.raise_for_status()
            return await response.text()

    async def set(self, key: str, value: Any) -> None:
        """Set a key in the database to the result of JSON encoding value.

        Args:
            key (str): The key to set
            value (Any): The value to set it to. Must be JSON-serializable.
        """
        await self.set_raw(key, _dumps(value))

    async def set_raw(self, key: str, value: str) -> None:
        """Set a key in the database to value.

        Args:
            key (str): The key to set
            value (str): The value to set it to
        """
        await self.set_bulk_raw({key: value})

    async def set_bulk(self, values: Dict[str, Any]) -> None:
        """Set multiple values in the database, JSON encoding them.

        Args:
            values (Dict[str, Any]): A dictionary of values to put into the dictionary.
                Values must be JSON serializeable.
        """
        await self.set_bulk_raw({k: _dumps(v) for k, v in values.items()})

    async def set_bulk_raw(self, values: Dict[str, str]) -> None:
        """Set multiple values in the database.

        Args:
            values (Dict[str, str]): The key-value pairs to set.
        """
        async with self.sess.post(self.db_url, data=values) as response:
            response.raise_for_status()

    async def delete(self, key: str) -> None:
        """Delete a key from the database.

        Args:
            key (str): The key to delete

        Raises:
            KeyError: Key does not exist
        """
        async with self.sess.delete(
            self.db_url + "/" + urllib.parse.quote(key)
        ) as response:
            if response.status == 404:
                raise KeyError(key)
            response.raise_for_status()

    async def list(self, prefix: str) -> Tuple[str, ...]:
        """List keys in the database which start with prefix.

        Args:
            prefix (str): The prefix keys must start with, blank not not check.

        Returns:
            Tuple[str]: The keys found.
        """
        params = {"prefix": prefix, "encode": "true"}
        async with self.sess.get(self.db_url, params=params) as response:
            response.raise_for_status()
            text = await response.text()
            if not text:
                return tuple()
            else:
                return tuple(urllib.parse.unquote(k) for k in text.split("\n"))

    async def to_dict(self, prefix: str = "") -> Dict[str, str]:
        """Dump all data in the database into a dictionary.

        Args:
            prefix (str): The prefix the keys must start with,
                blank means anything. Defaults to "".

        Returns:
            Dict[str, str]: All keys in the database.
        """
        ret = {}
        keys = await self.list(prefix=prefix)
        for i in keys:
            ret[i] = await self.get(i)
        return ret

    async def keys(self) -> Tuple[str, ...]:
        """Get all keys in the database.

        Returns:
            Tuple[str]: The keys in the database.
        """
        return await self.list("")

    async def values(self) -> Tuple[str, ...]:
        """Get every value in the database.

        Returns:
            Tuple[str]: The values in the database.
        """
        data = await self.to_dict()
        return tuple(data.values())

    async def items(self) -> Tuple[Tuple[str, str], ...]:
        """Convert the database to a dict and return the dict's items method.

        Returns:
            Tuple[Tuple[str]]: The items
        """
        return tuple((await self.to_dict()).items())

    def __repr__(self) -> str:
        """A representation of the database.

        Returns:
            A string representation of the database object.
        """
        return f"<{self.__class__.__name__}(db_url={self.db_url!r})>"


class ObservedList(abc.MutableSequence):
    """A list that calls a function every time it is mutated.

    Attributes:
        value (List): The underlying list.
    """

    __slots__ = ("_on_mutate_handler", "value")

    def __init__(
        self, on_mutate: Callable[[List], None], value: Optional[List] = None
    ) -> None:
        self._on_mutate_handler = on_mutate
        if value is None:
            self.value = []
        else:
            self.value = value

    def on_mutate(self) -> None:
        """Calls the mutation handler with the underlying list as an argument."""
        self._on_mutate_handler(self.value)

    def __getitem__(self, i: Union[int, slice]) -> Any:
        return self.value[i]

    def __setitem__(self, i: Union[int, slice], val: Any) -> None:
        self.value[i] = val
        self.on_mutate()

    def __delitem__(self, i: Union[int, slice]) -> None:
        del self.value[i]
        self.on_mutate()

    def __len__(self) -> int:
        return len(self.value)

    def __iter__(self) -> Iterator[Any]:
        return iter(self.value)

    def __imul__(self, rhs: Any) -> Any:
        self.value *= rhs
        self.on_mutate()
        return self.value

    def __eq__(self, rhs: Any) -> bool:
        return self.value == rhs

    def insert(self, i: int, elem: Any) -> None:
        """Inserts a value into the underlying list."""
        self.value.insert(i, elem)
        self.on_mutate()

    def set_value(self, value: List) -> None:
        """Sets the value attribute and triggers the mutation function."""
        self.value = value
        self.on_mutate()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(value={self.value!r})"


class ObservedDict(abc.MutableMapping):
    """A list that calls a function every time it is mutated.

    Attributes:
        value (Dict): The underlying dict.
    """

    __slots__ = ("_on_mutate_handler", "value")

    def __init__(
        self, on_mutate: Callable[[Dict], None], value: Optional[Dict] = None
    ) -> None:
        self._on_mutate_handler = on_mutate
        if value is None:
            self.value = {}
        else:
            self.value = value

    def on_mutate(self) -> None:
        """Calls the mutation handler with the underlying dict as an argument."""
        self._on_mutate_handler(self.value)

    def __contains__(self, k: Any) -> bool:
        return k in self.value

    def __getitem__(self, k: Any) -> Any:
        return self.value[k]

    # This should be posititional only but flake8 doesn't like that
    def get(self, key: str, default: Any = None) -> Any:
        """Return the value for key if key is in the dictionary, else default."""
        return self.value.get(
            key, item_to_observed(_get_set_cb(db=self, k=key), default)
        )

    def __setitem__(self, k: Any, v: Any) -> None:
        self.value[k] = v
        self.on_mutate()

    def __delitem__(self, k: Any) -> None:
        del self.value[k]
        self.on_mutate()

    def __iter__(self) -> Iterator[Any]:
        return iter(self.value)

    def __len__(self) -> int:
        return len(self.value)

    def __eq__(self, rhs: Any) -> bool:
        return self.value == rhs

    def __imul__(self, rhs: Any) -> Any:
        self.value *= rhs
        self.on_mutate()
        return self.value

    def set_value(self, value: Dict) -> None:
        """Sets the value attribute and triggers the mutation function."""
        self.value = value
        self.on_mutate()

    def __repr__(self) -> str:
        return f"{type(self).__name__}(value={self.value!r})"


# By putting these outside we save some memory
def _get_on_mutate_cb(d: Any) -> Callable[[Any], None]:
    def cb(_: Any) -> None:
        d.on_mutate()

    return cb


def _get_set_cb(db: Any, k: str) -> Callable[[Any], None]:
    def cb(val: Any) -> None:
        db[k] = val

    return cb


def item_to_observed(on_mutate: Callable[[Any], None], item: Any) -> Any:
    """Takes a JSON value and recursively converts it into an Observed value."""
    if isinstance(item, dict):
        # no-op handler so we don't call on_mutate in the loop below
        observed_dict = ObservedDict((lambda _: None), item)
        cb = _get_on_mutate_cb(observed_dict)

        for k, v in item.items():
            observed_dict[k] = item_to_observed(cb, v)

        observed_dict._on_mutate_handler = on_mutate
        return observed_dict
    elif isinstance(item, list):
        # no-op handler so we don't call on_mutate in the loop below
        observed_list = ObservedList((lambda _: None), item)
        cb = _get_on_mutate_cb(observed_list)

        for i, v in enumerate(item):
            observed_list[i] = item_to_observed(cb, v)

        observed_list._on_mutate_handler = on_mutate
        return observed_list
    else:
        return item


class Database(abc.MutableMapping):
    """Dictionary-like interface for Repl.it Database.

    This interface will coerce all values everything to and from JSON. If you
    don't want this, use AsyncDatabase instead.
    """

    __slots__ = ("db_url", "sess")

    def __init__(self, db_url: str) -> None:
        """Initialize database. You shouldn't have to do this manually.

        Args:
            db_url (str): Database url to use.
        """
        self.db_url = db_url
        self.sess = requests.Session()

    def __getitem__(self, key: str) -> Any:
        """Get the value of an item from the database.

        Will replace the mutable JSON types of dict and list with subclasses that
        enable nested setting. These classes will block to request the DB on every
        mutation, which can have performance implications. To disable this, use the
        `get_raw` method instead.

        This method will JSON decode the value. To disable this behavior, use the
        `get_raw` method instead.

        Args:
            key (str): The key to retreive

        Returns:
            Any: The value of the key
        """
        raw_val = self.get_raw(key)
        val = json.loads(raw_val)
        return item_to_observed(_get_set_cb(self, key), val)

    # This should be posititional only but flake8 doesn't like that
    def get(self, key: str, default: Any = None) -> Any:
        """Return the value for key if key is in the database, else default.

        Will replace the mutable JSON types of dict and list with subclasses that
        enable nested setting. These classes will block to request the DB on every
        mutation, which can have performance implications. To disable this, use the
        `get_raw` method instead.

        This method will JSON decode the value. To disable this behavior, use the
        `get_raw` method instead.

        Args:
            key (str): The key to retreive
            default (Any): The default to return if the key is not the database.
                Defaults to None.

        Returns:
            Any: The the value for key if key is in the database, else default.
        """
        return super().get(key, item_to_observed(_get_set_cb(self, key), default))

    def get_raw(self, key: str) -> str:
        """Look up the given key in the database and return the corresponding value.

        Args:
            key (str): The key to look up

        Raises:
            KeyError: The key is not in the database.

        Returns:
            str: The value of the key in the database.
        """
        r = self.sess.get(self.db_url + "/" + urllib.parse.quote(key))
        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()
        return r.text

    def __setitem__(self, key: str, value: Any) -> None:
        """Set a key in the database to the result of JSON encoding value.

        Args:
            key (str): The key to set
            value (Any): The value to set it to. Must be JSON-serializable.
        """
        self.set(key, value)

    def set(self, key: str, value: Any) -> None:
        """Set a key in the database to value, JSON encoding it.

        Args:
            key (str): The key to set
            value (Any): The value to set.
        """
        self.set_raw(key, _dumps(value))

    def set_raw(self, key: str, value: str) -> None:
        """Set a key in the database to value.

        Args:
            key (str): The key to set
            value (str): The value to set.
        """
        self.set_bulk_raw({key: value})

    def set_bulk(self, values: Dict[str, Any]) -> None:
        """Set multiple values in the database, JSON encoding them.

        Args:
            values (Dict[str, Any]): A dictionary of values to put into the dictionary.
                Values must be JSON serializeable.
        """
        self.set_bulk_raw({k: _dumps(v) for k, v in values.items()})

    def set_bulk_raw(self, values: Dict[str, str]) -> None:
        """Set multiple values in the database.

        Args:
            values (Dict[str, str]): The key-value pairs to set.
        """
        r = self.sess.post(self.db_url, data=values)
        r.raise_for_status()

    def __delitem__(self, key: str) -> None:
        """Delete a key from the database.

        Args:
            key (str): The key to delete

        Raises:
            KeyError: Key is not set
        """
        r = self.sess.delete(self.db_url + "/" + urllib.parse.quote(key))
        if r.status_code == 404:
            raise KeyError(key)

        r.raise_for_status()

    def __iter__(self) -> Iterator[str]:
        """Return an iterator for the database."""
        return iter(self.prefix(""))

    def __len__(self) -> int:
        """The number of keys in the database."""
        return len(self.prefix(""))

    def prefix(self, prefix: str) -> Tuple[str, ...]:
        """Return all of the keys in the database that begin with the prefix.

        Args:
            prefix (str): The prefix the keys must start with,
                blank means anything.

        Returns:
            Tuple[str]: The keys found.
        """
        r = self.sess.get(f"{self.db_url}", params={"prefix": prefix, "encode": "true"})
        r.raise_for_status()

        if not r.text:
            return tuple()
        else:
            return tuple(urllib.parse.unquote(k) for k in r.text.split("\n"))

    def keys(self) -> AbstractSet[str]:
        """Returns all of the keys in the database.

        Returns:
            List[str]: The keys.
        """
        # Rationale for this method:
        # This is implemented for free from our superclass using iter, but when you
        #  db.keys() in the console, you should see the keys immediately. Without this,
        #  it will just print an ugly repr that doesn't show the data within.
        # By implementing this method we get pretty output in the console when you
        #  type db.keys() in an interactive prompt.

        # TODO: Return a set from prefix since keys are guaranteed unique
        return set(self.prefix(""))

    def dumps(self, val: Any) -> str:
        """JSON encodes a value that can be a special DB object."""
        return _dumps(val)

    def __repr__(self) -> str:
        """A representation of the database.

        Returns:
            A string representation of the database object.
        """
        return f"<{self.__class__.__name__}(db_url={self.db_url!r})>"

    def close(self) -> None:
        """Closes the database client connection."""
        self.sess.close()
